from bulbs.api.permissions import CanEditCmsNotifications
from bulbs.cms_notifications.models import CmsNotification
from bulbs.cms_notifications.serializers import CmsNotificationSerializer
from rest_framework import viewsets


class CmsNotificationsViewSet(viewsets.ModelViewSet):

    queryset = CmsNotification.objects.all()
    serializer_class = CmsNotificationSerializer
    permission_classes = [CanEditCmsNotifications]


notifications_view = CmsNotificationsViewSet.as_view({
    'get': 'list',
    'post': 'create',
    'put': 'update',
    'delete': 'destroy'
})



from rest_framework import serializers

from bulbs.content.serializers import ContentSerializer

from .models import ContentList


class ContentListField(serializers.WritableField):
    def field_to_native(self, obj, field_name):
        data = []
        for content in obj:
            data.append(ContentSerializer(instance=content).data)

        return data

    def from_native(self, data):
        return [{"id": content_data["id"]} for content_data in data]


class ContentListSerializer(serializers.ModelSerializer):
    class Meta:
        model = ContentList
        exclude = ("data",)

    content = ContentListField(source="data")



from bulbs.cms_notifications.api import notifications_view
from django.conf import settings
from django.conf.urls import url, include
from django.conf import settings

from .views import api_v1_router, MeViewSet

urlpatterns = (
    url(r"^me/logout/?$", "django.contrib.auth.views.logout", name="logout"),
    url(r'^me/?$', MeViewSet.as_view({'get': 'retrieve'}), name='me'),
    url(r"^", include(api_v1_router.urls))       # noqa
)

if "bulbs.cms_notifications" in settings.INSTALLED_APPS:
    urlpatterns += (
        url(r"^notifications/(?P<pk>\d+)?", notifications_view, name="notifications"),
    )

if "bulbs.contributions" in settings.INSTALLED_APPS:
    urlpatterns += (
        url(r"^contributions/", include('bulbs.contributions.urls')),
    )



from django.db import models

from polymorphic import PolymorphicModel


class ContentListOperation(PolymorphicModel):

    class Meta:
        ordering = ["-when"]

    content_list = models.ForeignKey("promotion.ContentList", related_name="operations")
    when = models.DateTimeField()
    applied = models.BooleanField(default=False)

    def apply(self, data):
        raise NotImplemented()


class InsertOperation(ContentListOperation):

    index = models.IntegerField(default=0)
    content = models.ForeignKey("content.Content", related_name="+")
    lock = models.BooleanField(default=False)
    
    def apply(self, data):
        next = {
            "id": self.content.pk,
            "lock": self.lock
        }
        for i in range(self.index, min(len(data), 100)):
            if data[i].get("lock", False):
                continue
            next, data[i] = data[i], next  # Swap them
        data.append(next)
        return data


class ReplaceOperation(ContentListOperation):

    content = models.ForeignKey("content.Content", related_name="+")
    target = models.ForeignKey("content.Content", related_name="+")
    lock = models.BooleanField(default=False)

    def apply(self, data):
        replace = {
            "id": self.content.pk,
            "lock": self.lock
        }
        for index, item in enumerate(data):
            if item["id"] == self.target.pk:
                if item.get("lock", False):
                    raise Exception("That item is locked!")
                data[index] = replace
                break
        else:
            raise Exception("No content in list!")
        return data


class LockOperation(ContentListOperation):

    target = models.ForeignKey("content.Content", related_name="+")

    def apply(self, data):
        for index, item in enumerate(data):
            if item["id"] == self.target.pk:
                data[index]["lock"] = True
                break
        else:
            raise Exception("No content in list!")
        return data


class UnlockOperation(ContentListOperation):

    target = models.ForeignKey("content.Content", related_name="+")

    def apply(self, data):
        for index, item in enumerate(data):
            if item["id"] == self.target.pk:
                data[index]["lock"] = False
                break
        else:
            raise Exception("No content in list!")
        return data



from django.conf.urls import patterns, include, url

from bulbs.videos.api import router

urlpatterns = patterns('bulbs.videos.views',
    url(r'api/', include(router.urls)),
    url(r'videoAttrs\.js$', 'video_attrs'),
    url(r'notification$', 'notification')
)



import os
import copy
import requests
import json

from django.core.exceptions import ImproperlyConfigured
from django.core.urlresolvers import reverse
from django.conf import settings

from rest_framework import serializers
from rest_framework import fields
from rest_framework import viewsets
from rest_framework import decorators
from rest_framework import routers
from rest_framework.response import Response

from .models import Video
from bulbs.content.models import LogEntry


class JSONField(serializers.WritableField):
    def to_native(self, obj):
        if obj is None:
            return self.default
        return obj

    def from_native(self, data):
        if data is None:
            return self.default
        return data


class VideoSerializer(serializers.ModelSerializer):

    status = fields.Field(source='get_status_display')
    sources = JSONField(required=False)

    class Meta:
        model = Video
        exclude = ("data",)
        read_only_fields = ("id", "job_id")


class VideoViewSet(viewsets.ModelViewSet):

    model = Video
    serializer_class = VideoSerializer

    @decorators.action()
    def encode(self, request, pk=None):
        video = self.get_object()
        if 'name' in request.POST:
            video.name = request.POST.get('name')

        try:
            s3_path = os.path.join(
                settings.VIDEO_ENCODING['bucket'],
                settings.VIDEO_ENCODING['directory'],
                str(video.pk))
        except KeyError:
            raise ImproperlyConfigured("Video encoding settings aren't defined.")

        base_url = "s3://%s" % s3_path
        default_notification_url = request.build_absolute_uri(
            reverse('bulbs.videos.views.notification'))

        payload = {
            'input': '%s/original' % base_url,
            'outputs': [],
            'notifications': [{
                "url": settings.VIDEO_ENCODING.get("notification_url", default_notification_url),
                "format": "json"
            }]
        }

        output_update = {
            'base_url': base_url,
            'public': True
        }
        for output_template in settings.VIDEO_ENCODING.get('outputs', []):
            output = copy.copy(output_template)
            if "thumbnails" in output_template:
                output["thumbnails"].update(output_update)
            else:
                output.update(output_update)
            payload['outputs'].append(output)

        auth_headers = {'Zencoder-Api-Key': settings.VIDEO_ENCODING.get('zencoder_api_key')}
        response = requests.post(
            "https://app.zencoder.com/api/v2/jobs",
            data=json.dumps(payload), headers=auth_headers)
        if response.status_code == 201:
            video.job_id = response.json().get('id')
            video.status = Video.IN_PROGRESS
            video.data = response.json()
            video.save()
            if request.user.is_authenticated():
                LogEntry.objects.log(request.user, video, "Encoded")
            else:
                LogEntry.objects.log(None, video, "Encoded")

        return Response(response.json(), status=response.status_code)

    def post_save(self, obj, created=False):
        message = "Created" if created else "Saved"
        if self.request.user.is_authenticated():
            LogEntry.objects.log(self.request.user, obj, message)
        else:
            LogEntry.objects.log(None, obj, message)
        return super(VideoViewSet, self).post_save(obj, created=created)

router = routers.SimpleRouter()
router.register('video', VideoViewSet)



from django import template

register = template.Library()


@register.inclusion_tag('videos/player.html')
def player(video, width=None, height=None):
    context = {'video': video}
    if width:
        context['width'] = width
        if height is None:
            context['height'] = int(width * 9 / 16)
    if height:
        context['height'] = height
        if width is None:
            context['width'] = int(width * 16 / 9)
    return context



from django.db import models

from django.contrib.auth import get_user_model
User = get_user_model()

from bulbs.content.models import Content


class ContributorRole(models.Model):
    name = models.CharField(max_length=255)
    description = models.TextField(null=True, blank=True)


class Contribution(models.Model):
    role = models.ForeignKey(ContributorRole)
    contributor = models.ForeignKey(User)
    content = models.ForeignKey(Content)
    notes = models.TextField(null=True, blank=True)



"""Base models for "Content", including the indexing and search features
that we want any piece of content to have."""

from django.conf import settings
from django.contrib.contenttypes.models import ContentType
from django.core.urlresolvers import NoReverseMatch, reverse
from django.db import models
from django.template.defaultfilters import slugify
from django.utils import timezone
from django.utils.html import strip_tags

from bulbs.content import TagCache
import elasticsearch
from elasticutils import F
from elastimorphic.base import (
    PolymorphicIndexable,
    PolymorphicMappingType,
    SearchManager,
)
from polymorphic import PolymorphicModel

from .shallow import ShallowContentS, ShallowContentResult

from djbetty import ImageField

try:
    from bulbs.content.tasks import index as index_task  # noqa
    from bulbs.content.tasks import update as update_task  # noqa
    CELERY_ENABLED = True
except ImportError:
    CELERY_ENABLED = False


class Tag(PolymorphicIndexable, PolymorphicModel):
    """Model for tagging up Content."""
    name = models.CharField(max_length=255)
    slug = models.SlugField(unique=True)

    search_objects = SearchManager()

    def __unicode__(self):
        return '%s: %s' % (self.__class__.__name__, self.name)

    def save(self, *args, **kwargs):
        self.slug = slugify(self.name)
        return super(Tag, self).save(*args, **kwargs)

    def count(self):
        return TagCache.count(self.slug)

    @classmethod
    def get_mapping_properties(cls):
        props = super(Tag, cls).get_mapping_properties()
        props.update({
            "name": {"type": "string", "analyzer": "autocomplete"},
            "slug": {"type": "string", "index": "not_analyzed"},
            "type": {"type": "string", "index": "not_analyzed"}
        })
        return props

    def extract_document(self):
        data = super(Tag, self).extract_document()
        data.update({
            "name": self.name,
            "slug": self.slug,
            "type": self.get_mapping_type_name()
        })
        return data

    @classmethod
    def get_serializer_class(cls):
        from .serializers import TagSerializer
        return TagSerializer


class FeatureType(models.Model):
    name = models.CharField(max_length=255)
    slug = models.SlugField(unique=True)

    def __unicode__(self):
        return self.name

    def save(self, *args, **kwargs):
        if self.slug is None or self.slug == "":
            self.slug = slugify(self.name)
        return super(FeatureType, self).save(*args, **kwargs)


class ContentManager(SearchManager):

    def s(self):
        """Returns a ShallowContentS() instance, using an ES URL from the settings, and an index
        from this manager's model"""

        base_polymorphic_class = self.model.get_base_class()
        type_ = type(
            '%sMappingType' % base_polymorphic_class.__name__,
            (PolymorphicMappingType,),
            {"base_polymorphic_class": base_polymorphic_class})

        return ShallowContentS(type_=type_).es(urls=settings.ES_URLS)

    def search(self, **kwargs):
        """
        Queries using ElasticSearch, returning an elasticutils queryset.

        Allowed params:

         * query
         * tags
         * types
         * feature_types
         * authors
         * published
        """

        results = self.s()

        if "query" in kwargs:
            results = results.query(_all__match=kwargs.get("query"))
        else:
            results = results.order_by('-published', '-last_modified')

        # Right now we have "Before", "After" (datetimes),
        # and "published" (a boolean). Should simplify this in the future.
        if "before" in kwargs or "after" in kwargs:
            if "before" in kwargs:
                results = results.query(published__lte=kwargs["before"], must=True)

            if "after" in kwargs:
                results = results.query(published__gte=kwargs["after"], must=True)
        else:
            if kwargs.get("published", True) and not "status" in kwargs:  # TODO: kill this "published" param. it sucks
                now = timezone.now()
                results = results.query(published__lte=now, must=True)

        if "status" in kwargs:
            results = results.filter(status=kwargs.get("status"))

        f = F()
        for tag in kwargs.get("tags", []):
            if tag.startswith("-"):
                f &= ~F(**{"tags.slug": tag[1:]})
            else:
                f |= F(**{"tags.slug": tag})

        for feature_type in kwargs.get("feature_types", []):
            if feature_type.startswith("-"):
                f &= ~F(**{"feature_type.slug": feature_type[1:]})
            else:
                f |= F(**{"feature_type.slug": feature_type})

        for author in kwargs.get("authors", []):
            if author.startswith("-"):
                f &= ~F(**{"authors.username": author})
            else:
                f |= F(**{"authors.username": author})

        results = results.filter(f)

        # only use valid subtypes
        types = kwargs.pop("types", [])
        model_types = self.model.get_mapping_type_names()
        if types:
            results = results.doctypes(*[
                type_classname for type_classname in types \
                if type_classname in model_types
            ])
        else:
            results = results.doctypes(*model_types)
        return results

    def in_bulk(self, pks):
        results = self.es.multi_get(pks, index=self.model.get_index_name())
        ret = []
        for r in results["docs"]:
            if "_source" in r:
                ret.append(ShallowContentResult(r["_source"]))
            else:
                ret.append(None)
        return ret


class Content(PolymorphicIndexable, PolymorphicModel):
    """The base content model from which all other content derives."""
    published = models.DateTimeField(blank=True, null=True)
    last_modified = models.DateTimeField(auto_now=True, default=timezone.now)
    title = models.CharField(max_length=512)
    slug = models.SlugField(blank=True, default='')
    description = models.TextField(max_length=1024, blank=True, default="")

    # field used if thumbnail has been manually overridden
    thumbnail_override = ImageField(null=True, blank=True, editable=False)

    authors = models.ManyToManyField(settings.AUTH_USER_MODEL)
    feature_type = models.ForeignKey(FeatureType, null=True, blank=True)
    subhead = models.CharField(max_length=255, blank=True, default="")

    tags = models.ManyToManyField(Tag, blank=True)

    indexed = models.BooleanField(default=True)  # Should this item be indexed?

    _readonly = False  # Is this a read only model? (i.e. from elasticsearch)

    search_objects = ContentManager()

    class Meta:
        permissions = (
            ("publish_own_content", "Can publish their own content"),
            ("publish_content", "Can publish content"),
            ("promote_content", "Can promote content"),
        )

    def __unicode__(self):
        return '%s: %s' % (self.__class__.__name__, self.title)

    @property
    def thumbnail(self):
        """Read-only attribute that provides the value of the thumbnail to display."""

        # check if there is a valid thumbnail override
        if self.thumbnail_override.id is not None:
            return self.thumbnail_override

        # otherwise, just try to grab the first image
        first_image = self.first_image
        if first_image is not None:
            return first_image

        # no override for thumbnail and no non-none image field, just return override, which is a blank image field
        return self.thumbnail_override

    @property
    def first_image(self):
        """Ready-only attribute that provides the value of the first non-none image that's not the thumbnail override
        field."""

        # loop through image fields and grab the first non-none one
        for field in self._meta.fields:
            if isinstance(field, ImageField):
                if field.name is not 'thumbnail_override':
                    field_value = getattr(self, field.name)
                    if field_value.id is not None:
                        return field_value

        # no non-none images, return None
        return None

    def get_absolute_url(self):
        try:
            url = reverse("content-detail-view", kwargs={"pk": self.pk, "slug": self.slug})
        except NoReverseMatch:
            url = None
        return url

    def get_status(self):
        """Returns a string representing the status of this item

        By default, this is one of "draft", "scheduled" or "published"."""

        if self.published:
            return "final"  # The published time has been set

        return "draft"  # No published time has been set

    @property
    def is_published(self):
        if self.published:
            now = timezone.now()
            if now >= self.published:
                return True
        return False

    @property
    def type(self):
        return self.get_mapping_type_name()

    def ordered_tags(self):
        tags = list(self.tags.all())
        return sorted(
            tags, key=lambda tag: ((type(tag) != Tag) * 100000) + tag.count(), reverse=True)

    def build_slug(self):
        return strip_tags(self.title)

    def save(self, *args, **kwargs):
        if not self.slug:
            self.slug = slugify(self.build_slug())[:self._meta.get_field("slug").max_length]
        if self.indexed is False:
            if kwargs is None:
                kwargs = {}
            kwargs["index"] = False
        return super(Content, self).save(*args, **kwargs)

    # class methods ##############################
    @classmethod
    def get_mapping_properties(cls):
        properties = super(Content, cls).get_mapping_properties()
        properties.update({
            "published": {"type": "date"},
            "last_modified": {"type": "date"},
            "title": {"type": "string", "analyzer": "snowball", "_boost": 2.0},
            "slug": {"type": "string"},
            "description": {"type": "string", },
            "thumbnail": {"type": "integer"},
            "feature_type": {
                "properties": {
                    "name": {
                        "type": "multi_field",
                        "fields": {
                            "name": {"type": "string", "index": "not_analyzed"},
                            "autocomplete": {"type": "string", "analyzer": "autocomplete"}
                        }
                    },
                    "slug": {"type": "string", "index": "not_analyzed"}
                }
            },
            "authors": {
                "properties": {
                    "first_name": {"type": "string"},
                    "id": {"type": "long"},
                    "last_name": {"type": "string"},
                    "username": {"type": "string", "index": "not_analyzed"}
                }
            },
            "tags": {
                "properties": Tag.get_mapping_properties()
            },
            "absolute_url": {"type": "string"},
            "status": {"type": "string", "index": "not_analyzed"}
        })
        return properties

    def extract_document(self):
        data = super(Content, self).extract_document()
        data.update({
            "published"        : self.published,
            "last_modified"    : self.last_modified,
            "title"            : self.title,
            "slug"             : self.slug,
            "description"      : self.description,
            "thumbnail"        : self.thumbnail.id if self.thumbnail else None,
            "authors": [{
                "first_name": author.first_name,
                "id"        : author.id,
                "last_name" : author.last_name,
                "username"  : author.username
            } for author in self.authors.all()],
            "tags": [tag.extract_document() for tag in self.ordered_tags()],
            "absolute_url": self.get_absolute_url(),
            "status": self.get_status()
        })
        if self.feature_type:
            data["feature_type"] = {
                "name": self.feature_type.name,
                "slug": self.feature_type.slug
            }
        return data

    @classmethod
    def get_serializer_class(cls):
        from .serializers import ContentSerializer
        return ContentSerializer


class LogEntryManager(models.Manager):
    def log(self, user, content, message):
        return self.create(
            user=user,
            content_type=ContentType.objects.get_for_model(content),
            object_id=content.pk,
            change_message=message
        )


class LogEntry(models.Model):
    action_time = models.DateTimeField("action time", auto_now=True)
    content_type = models.ForeignKey(ContentType, blank=True, null=True, related_name="change_logs")
    object_id = models.TextField(("object id"), blank=True, null=True)
    user = models.ForeignKey(settings.AUTH_USER_MODEL, null=True, blank=True, related_name="change_logs")
    change_message = models.TextField("change message", blank=True)

    objects = LogEntryManager()

    class Meta:
        ordering = ("-action_time",)


def content_deleted(sender, instance=None, **kwargs):
    if getattr(instance, "_index", True):
        index = instance.get_index_name()
        klass = instance.get_real_instance_class()
        try:
            klass.search_objects.es.delete(index, klass.get_mapping_type_name(), instance.id)
        except elasticsearch.exceptions.NotFoundError:
            pass

models.signals.pre_delete.connect(content_deleted, Content)



from bulbs.cms_notifications.models import CmsNotification
from rest_framework import serializers


class CmsNotificationSerializer(serializers.ModelSerializer):

    post_date = serializers.DateTimeField()
    notify_end_date = serializers.DateTimeField()

    def validate(self, attrs):
        """Ensure that post_date occurs before notify_end_date."""

        if attrs['post_date'] > attrs['notify_end_date']:
            raise serializers.ValidationError("Post date must occur before promotion end date.")

        return attrs

    class Meta:
        model = CmsNotification



from rest_framework import permissions


class HasPermissionOrIsAuthor(permissions.BasePermission):
    authors_field = "authors"
    require_staff = True
    permissions = []  # permissions which would permit a non-author to do this
    author_permissions = []  # permissions the author needs in order to be permitted
    protected_methods = []

    def has_object_permission(self, request, view, obj):
        # Give permission if we're not protecting this method
        if self.protected_methods and not request.method in self.protected_methods:
            return True
        user = getattr(request, "user", None)
        if not user or user.is_anonymous():
            return False
        if self.require_staff and not user.is_staff:
            return False
        # if they have higher-level privileges we can return true right now
        if user.has_perms(self.permissions):
            return True
        # no? ok maybe they're the author and have appropriate author permissions.
        authors_field = getattr(obj, self.authors_field, None)
        if not authors_field:
            return False
        if self.author_permissions and not user.has_perms(self.author_permissions):
            return False
        return user in authors_field.all()


class CanEditContent(HasPermissionOrIsAuthor):
    permissions = ["content.change_content"]
    protected_methods = ["PUT"]


class CanPromoteContent(HasPermissionOrIsAuthor):
    permissions = ["content.promote_content"]


class CanPublishContent(HasPermissionOrIsAuthor):
    permissions = ["content.publish_content"]
    author_permissions = ["content.publish_own_content"]


class CanEditCmsNotifications(permissions.BasePermission):

    def has_permission(self, request, view):
        """If method is GET, user can access, if method is PUT or POST user must be a superuser."""
        has_permission = False

        if request.method == "GET" \
                or request.method in ["PUT", "POST", "DELETE"] and request.user and request.user.is_superuser:
            has_permission = True

        return has_permission


from django.db import models

from json_field import JSONField


DEFAULT_VIDEO_OUTPUT = {
    "public": True,
    "width": 640,
    "height": 320
}


VIDEO_PREFERENCES = {
    "video/mp4" : 0.75,
    "video/webm": 1.25
}


class Video(models.Model):
    """This is a very lightweight model that basically wraps an externally available set of sources
    for a given video."""

    NOT_STARTED = 0
    COMPLETE = 1
    IN_PROGRESS = 2
    FAILED = 3
    STATUSES = (
        (NOT_STARTED, 'Not started'),
        (COMPLETE, 'Complete'),
        (IN_PROGRESS, 'In Progress'),
        (FAILED, 'Failed')
    )

    name = models.CharField(max_length=255)
    data = JSONField(null=True, blank=True, help_text="This is JSON that is returned from an encoding job")
    sources = JSONField(null=True, blank=True, default=[], help_text="This is a JSON array of sources.")
    poster_url = models.URLField(null=True, blank=True)
    status = models.IntegerField(choices=STATUSES, default=NOT_STARTED)
    job_id = models.IntegerField(null=True, blank=True, help_text="The zencoder job ID")

    def __unicode__(self):
        return self.name

    def ordered_sources(self):
        return sorted(
            self.sources,
            key=lambda source: VIDEO_PREFERENCES.get(
                source['content_type'], 1
            ) * source.get('width', 1))
